/*
Copyright (C) 2018 Falcury

DESCRIPTION:
    Program for playing back Prince of Persia MIDI files.

CREDITS:
    The OPL integration code is based in part on Chocolate Doom's MIDI interface, by Simon Howard (GPLv2-licensed).
    Uses the Nuked OPL3 emulator by Alexey Khokholov (GPLv2-licensed).

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "popmidi.h"
#include "opl3.h"

bool32 is_playing;

void sdl_panic(const char* header) {
	const char* error = SDL_GetError();
	printf("%s: %s\n", header, error);
	panic_and_exit();
}

i32 audio_frames_left;
f64 t_sine;
f64 time_division;
f64 sine_frequency;
#define MIX_BUFFER_SIZE ((u32)MEGABYTES(1))
u8* mix_buffer;
u32 mix_position_offset;

opl3_chip opl_chip;
instrument_type* instruments;
u32 num_instruments;
int ticks_per_beat;
int us_per_beat;
int mixing_freq;
u8 voice_note[9];
int voice_instrument[9];
int channel_instrument[16];

instrument_type hardcoded_instrument = {
		0x13, 0x09, 0x04, {{0x02, 0x8D, 0xD7, 0x37, 0x00}, {0x03, 0x03, 0xF5, 0x18, 0x00}}, 0x00, 0x00, 0x00
};
parsed_midi_type parsed_midi;
midi_state_type midi_state;


static inline u8* get_mix_buffer_position() {
	u8* result = (mix_buffer + (mix_position_offset % MIX_BUFFER_SIZE));
	return result;
}

#pragma pack(push, 1)
typedef struct {i16 left, right;} stereo_pair;
#pragma pack(pop)

// Output a simple sine wave. (Unused - just messing around with audio callbacks, and filling a circular buffer)
void sine_audio_callback(void* userdata, Uint8* stream, int len) {
	i32 num_frames = len / 4;
	SDL_memset(stream, 0, len);
	if (audio_frames_left <= 0) {
		is_playing = 0;
	} else {
		u8* mix_buffer_pos = get_mix_buffer_position();
		u32 end_offset = (mix_position_offset % MIX_BUFFER_SIZE) + len;
		u32 bytes_to_write = len;
		u32 wrapped_bytes = 0;
		if (end_offset > MIX_BUFFER_SIZE) {
			// mix buffer is overflowing, so we'll need to wrap around to the beginning.
			u32 overflow = end_offset - MIX_BUFFER_SIZE;
			bytes_to_write -= overflow;
			wrapped_bytes = overflow;
		}

		for (int i = 0; i < num_frames; ++i) {
			if (t_sine > two_pi64) t_sine -= two_pi64;
			f64 sine_value = sin(t_sine * sine_frequency) * 0.1;
			i16 sine_value16 = (i16)(sine_value * INT16_MAX);
//			printf("%d\t", sine_value16);
			stereo_pair* pair = (stereo_pair*) get_mix_buffer_position();
			pair->left = sine_value16;
			pair->right= sine_value16;
			mix_position_offset += 4;
			t_sine += time_division;
		}
		SDL_memcpy(stream, mix_buffer_pos, bytes_to_write);
		if (wrapped_bytes != 0) {
			// get the remaining bytes from the beginning of the (circular) mix buffer.
			SDL_memcpy(stream + bytes_to_write, mix_buffer, wrapped_bytes);
		}
		audio_frames_left -= num_frames;
	}
//	printf("Audio callback was called...\n");
}

// Read a variable length integer (max 4 bytes).
static u32 midi_read_variable_length(u8** buffer_position) {
	u32 result = 0;
	u8* pos = *buffer_position;
	int i;
	for (i = 0; i < 4; ++i) {
		result = (result << 7) | (pos[i] & 0x7F);
		if ((pos[i] & 0x80) == 0) break; // The most significant bit being 0 means that this is the last u8.
	}
	*buffer_position += i+1; // Advance the pointer, so we know where the next field starts.
	return result;
}

bool32 parse_midi(midi_raw_chunk_type* midi, parsed_midi_type* parsed_midi) {
	ticks_per_beat = 24;
	if (memcmp(midi->chunk_type, "MThd", 4) != 0) {
		printf("Warning: Tried to play a midi sound without the 'MThd' chunk header.\n");
		return 0;
	}
	if (SDL_SwapBE32(midi->chunk_length) != 6) {
		printf("Warning: Midi file with an invalid header length (expected 6, is %d)\n",
		       SDL_SwapBE32(midi->chunk_length));
		return 0;
	}
	u16 midi_format = SDL_SwapBE16(midi->header.format);
	if (midi_format >= 2) {
		printf("Warning: Unsupported midi format %d (only type 0 or 1 files are supported)\n", midi_format);
		return 0;
	}
	u16 num_tracks = SDL_SwapBE16(midi->header.num_tracks);
	if (num_tracks < 1) {
		printf("Warning: Midi sound does not have any tracks.\n");
		return 0;
	}
	int division = SDL_SwapBE16(midi->header.time_division);
	if (division < 0) {
		division = (-(division / 256)) * (division & 0xFF); // Translate time delta from the alternative SMTPE format.
	}
	midi_state.time_division = division;
	ticks_per_beat = division;

	parsed_midi->tracks = calloc(1, num_tracks * sizeof(midi_track_type));
	parsed_midi->num_tracks = num_tracks;
	midi_raw_chunk_type* next_track_chunk = (midi_raw_chunk_type*) midi->header.tracks; // The first track chunk starts after the header chunk.
	u8 last_event_type = 0;
	for (int i = 0; i < num_tracks; ++i) {
		midi_raw_chunk_type* track_chunk = next_track_chunk;
		if (memcmp(track_chunk->chunk_type, "MTrk", 4) != 0) {
			printf("Warning: midi track without 'MTrk' chunk header.\n");
			free(parsed_midi->tracks);
			memset(&parsed_midi, 0, sizeof(parsed_midi));
			return 0;
		}
		next_track_chunk = (void*) track_chunk->data + (u32) SDL_SwapBE32(track_chunk->chunk_length);
		midi_track_type* track = &parsed_midi->tracks[i];
		u8* buffer_position = track_chunk->data;
		for (;;) {
			track->events = realloc(track->events, (++track->num_events) * sizeof(midi_event_type));
			midi_event_type* event = &track->events[track->num_events - 1];
			event->delta_time = midi_read_variable_length(&buffer_position);
			event->event_type = *buffer_position;
			if (event->event_type & 0x80) {
				last_event_type = event->event_type;
				++buffer_position;
			} else {
				event->event_type = last_event_type; // Implicit use of the previous event type.
			}
			// Determine the event type and parse the event.
			int num_channel_event_params = 1;
			switch (event->event_type & 0xF0) {
				case 0x80: // note off
				case 0x90: // note on
				case 0xA0: // aftertouch
				case 0xB0: // controller
				case 0xE0: // pitch bend
					num_channel_event_params = 2; //fallthrough
				case 0xC0: // program change
				case 0xD0: { // channel aftertouch
					// Read the channel event.
					event->channel.channel = event->event_type & 0x0F;
					event->event_type &= 0xF0;
					event->channel.param1 = *buffer_position++;
					if (num_channel_event_params == 2) {
						event->channel.param2 = *buffer_position++;
					}
				}
					break;
				default:
					// Not a channel event.
					switch (event->event_type) {
						case 0xF0: // SysEx
						case 0xF7: // SysEx split
							// Read SysEx event
							event->sysex.length = midi_read_variable_length(&buffer_position);
							event->sysex.data = buffer_position;
							buffer_position += event->sysex.length;
							break;
						case 0xFF: // Meta event
							event->meta.type = *buffer_position++;
							event->meta.length = midi_read_variable_length(&buffer_position);
							event->meta.data = buffer_position;
							buffer_position += event->meta.length;
							break;
						default:
							printf("Warning: unknown midi event type 0x%02x (track %d, event %d)\n",
							       event->event_type, i, track->num_events - 1);
							free(parsed_midi->tracks);
							memset(&parsed_midi, 0, sizeof(parsed_midi));
							return 0;
					}
			}
			if (event->event_type == 0xFF /* meta event */ && event->meta.type == 0x2F /* end of track */) {
				break;
			}

		}

	}

//	printf("Midi file looks good...\n");
	return 1;
}

void print_midi_event(midi_event_type* event) {
	switch (event->event_type) {
		case 0x80: // note off
			printf("noteoff: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;
		case 0x90: // note on
			printf("noteon: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			{
				float octaves_from_A4 = ((int)event->channel.param1 - 69) / 12.0f;
				float frequency = powf(2.0f,  octaves_from_A4) * 440.0f;
				float f_number_float = frequency * (float)(1 << 20) / 49716.0f;
				int b = (int)(log2f(f_number_float) - 9) & 7;
				int f = ((int)f_number_float >> b) & 1023;
				printf(", frequency = %.1f Hz, F=%d, b=%d", frequency, f, b);
			}
			break;
		case 0xA0: // aftertouch
			printf("aftertouch: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;
		case 0xB0: // controller
			printf("controller: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;
		case 0xE0: // pitch bend
			printf("pitch bend: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;
		case 0xC0: // program change
			printf("program change: ch %d, par %x", event->channel.channel, event->channel.param1);
			break;
		case 0xD0:
			printf("channel aftertouch: ch %d, par %x", event->channel.channel, event->channel.param1);
			break;
		case 0xF0: // SysEx
			printf("sysex event (length=%d): ", event->sysex.length);
			for (int i = 0; i<event->sysex.length; ++i) {
				printf("%02x ", event->sysex.data[i]);
			}
			break;
		case 0xF7: // SysEx split
			printf("sysex split event");
			// Read SysEx event
			break;
		case 0xFF: // Meta event
			printf("meta event (length=%d): ", event->meta.length);
			switch(event->meta.type) {
				default:
					printf("unknown type (%x)", event->meta.type);
					break;
				case 0:
					printf("sequence number");
					break;
				case 1:
					printf("text event");
					break;
				case 2:
					printf("copyright notice");
					break;
				case 3:
					printf("sequence/track name: ");
					{
						char* text = malloc(event->meta.length+1);
						memcpy(text, event->meta.data, event->meta.length);
						text[event->meta.length] = '\0';
						printf("%s", text);
						free(text);
					}
					break;
				case 4:
					printf("instrument name");
					break;
				case 0x51:
					printf("set tempo: ");
					{
						u8* data = event->meta.data;
						int new_tempo = (data[0]<<16) | (data[1]<<8) | (data[2]);
						printf("set tempo: %d", new_tempo);
					}
					break;
				case 0x54:
					printf("SMTPE offset: ");
					for (int i = 0; i<event->meta.length; ++i) {
						printf("%02x ", event->meta.data[i]);
					}
					break;
				case 0x58:
					printf("time signature: ");
					for (int i = 0; i<event->meta.length; ++i) {
						printf("%02x ", event->meta.data[i]);
					}
					break;
				case 0x2F:
					printf("end of track");
					break;
			}
			break;
		default: break;

	}
	putchar('\n');
}

void print_midi_events(parsed_midi_type* parsed_midi) {
	for (int track_index = 0; track_index < parsed_midi->num_tracks; ++track_index) {
		midi_track_type* track = &parsed_midi->tracks[track_index];
		for (int event_index = 0; event_index < track->num_events; ++event_index) {
			midi_event_type* event = &track->events[event_index];
			printf("Track %d: Event %d (dt=%d): type = %x; ", track_index, event_index, event->delta_time, event->event_type);
			print_midi_event(event);
		}
	}
}

file_buffer_type* load_file_into_memory(const char* filename) {
	FILE* fp = fopen(filename, "rb");
	if (!fp) {
		printf("Couldn't open file '%s'.\n", filename);
		return NULL;
	}
	fseek(fp, 0, SEEK_END);
	i32 filesize = ftell(fp);
	fseek(fp, 0, SEEK_SET);
	if (filesize <= 0) {
		printf("Error reading file '%s'.\n", filename);
		return NULL;
	}
	file_buffer_type* result = malloc((size_t)(filesize + sizeof(((file_buffer_type*)0)->size)));
	fread(result->data, 1, (size_t)filesize, fp);
	fclose(fp);
	return result;
}

u8 opl_cached_regs[256];

static void opl_reset(int freq) {
	OPL3_Reset(&opl_chip, freq);
	memset(opl_cached_regs, 0, sizeof(opl_cached_regs));
}

static void opl_write_reg(u8 reg, u8 value) {
	OPL3_WriteReg(&opl_chip, reg, value);
	opl_cached_regs[reg] = value;
}

static void opl_write_reg_masked(u8 reg, u8 value, u8 mask) {
	u8 cached = opl_cached_regs[reg] & ~mask;
	value = cached | (value & mask);
	opl_write_reg(reg, value);
}

static const int voice_operators[2][9] = {
		{ 0x00, 0x01, 0x02, 0x08, 0x09, 0x0a, 0x10, 0x11, 0x12 },
		{ 0x03, 0x04, 0x05, 0x0b, 0x0c, 0x0d, 0x13, 0x14, 0x15 }
};

void opl_write_instrument(instrument_type* instrument, u8 voice) {
	opl_write_reg(0xC0 + voice, instrument->FB_conn);
	for (int operator_index = 0; operator_index < 2; ++operator_index) {
		operator_type* operator = &instrument->operators[operator_index];
		int op_reg = voice_operators[operator_index][voice];
		opl_write_reg(0x20 + op_reg, operator->mul);
		opl_write_reg(0x40 + op_reg, operator->ksl_tl);
		opl_write_reg(0x60 + op_reg, operator->a_d);
		opl_write_reg(0x80 + op_reg, operator->s_r);
		opl_write_reg(0xE0 + op_reg, operator->waveform);
	}

}

void midi_note_off(midi_event_type* event) {
	u8 note = event->channel.param1;
	u8 channel = event->channel.channel;
	for (int i = 0; i < 9; ++i) {
		if (voice_instrument[i] == channel_instrument[channel] && voice_note[i] == note) {
			opl_write_reg(0xB0 + i, 0);
			voice_note[i] = 0; // This voice is now free to be re-used by a new note.
			break;
		}
	}
}

void midi_note_on(midi_event_type* event) {
	u8 note = event->channel.param1;
	u8 velocity = event->channel.param2;
	u8 channel = event->channel.channel;

	if (velocity == 0) {
		midi_note_off(event);
	} else {
		// Find a free OPL voice.
		int voice = 0;
		for (int i = 0; i < 9; ++i) {
			if (voice_note[i] == 0) {
				voice = i;
				break;
			}
		}
		// Set the correct instrument for this voice.
		if (voice_instrument[voice] != channel_instrument[channel]) {
			opl_write_instrument(&instruments[channel_instrument[channel]], voice);
			voice_instrument[voice] = channel_instrument[channel];
		}
		voice_note[voice] = note;
		// Calculate frequency for a MIDI note: note number 69 = A4 = 440 Hz.
		float octaves_from_A4 = ((int)event->channel.param1 - 69) / 12.0f;
		float frequency = powf(2.0f,  octaves_from_A4) * 440.0f;
		float f_number_float = frequency * (float)(1 << 20) / 49716.0f;
		int block = (int)(log2f(f_number_float) - 9) & 7;
		int f = ((int)f_number_float >> block) & 1023;
//		opl_write_reg(0xB0 + voice, 0);
		opl_write_reg(0xA0 + voice, f & 0xFF);
		opl_write_reg(0xB0 + voice, 0x20 | (block << 2) | (f >> 8));

		// TODO: what is being done with the 'velocity' MIDI parameter?? It is being ignored, or...?
//		u8 op0 = voice_operators[0][voice];
//		u8 op1 = voice_operators[1][voice];
//		u8 volume = 0x3F - (volume_mapping_table[event->channel.param2] >> 2);
//		opl_write_reg_masked(0x40 + op0, event->channel.param2, 0x3F); //volume is ignored?
//		opl_write_reg_masked(0x40 + op1, event->channel.param2, 0x3F);
	}

}

void process_midi_event(midi_event_type* event) {
	print_midi_event(event);
	switch (event->event_type) {
		case 0x80: // note off
			midi_note_off(event);
			break;
		case 0x90: // note on
			midi_note_on(event);
			break;
		case 0xC0: // program change
			channel_instrument[event->channel.channel] = event->channel.param1;
			break;
		case 0xFF: // Meta event
			switch(event->meta.type) {
				default: break;
				case 0x51: // set tempo
					{
						u8* data = event->meta.data;
						int new_tempo = (data[0]<<16) | (data[1]<<8) | (data[2]);
						us_per_beat = new_tempo;
					}
					break;
				case 0x54: // SMTPE offset
					break;
				case 0x58: // time signature
					break;
				case 0x2F: // end of track
					break;
			}
			break;
#if 0
		case 0xA0: // aftertouch
			printf("aftertouch: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;
		case 0xB0: // controller
			printf("controller: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;
		case 0xE0: // pitch bend
			printf("pitch bend: ch %d, par %x|%x", event->channel.channel, event->channel.param1, event->channel.param2);
			break;

		case 0xD0:
			printf("channel aftertouch: ch %d, par %x", event->channel.channel, event->channel.param1);
			break;
		case 0xF0: // SysEx
			printf("sysex event (length=%d): ", event->sysex.length);
			for (int i = 0; i<event->sysex.length; ++i) {
				printf("%02x \n", event->sysex.data[i]);
			}
			break;
		case 0xF7: // SysEx split
			printf("sysex split event");
			// Read SysEx event
			break;

#endif
		default: break;

	}

}

#define ONE_SECOND_IN_US 1000000

void opl_audio_callback(void* userdata, Uint8* stream, int len) {
	if (len <= 0) return;
	int frames_needed = len / 4;
	while (frames_needed > 0) {
		if (midi_state.time_to_next_pause > 0) {
			// Fill the audio buffer (we have already processed the MIDI events up till this point)
			int available_frames = ((midi_state.time_to_next_pause * mixing_freq) + ONE_SECOND_IN_US - 1) / ONE_SECOND_IN_US; // round up.
			int advance_frames = MIN(available_frames, frames_needed);
			OPL3_GenerateStream(&opl_chip, (i16*)stream, advance_frames);
			frames_needed -= advance_frames;
			stream += advance_frames * 4;
			i64 us_elapsed = (advance_frames * ONE_SECOND_IN_US) / mixing_freq;
			midi_state.time_to_next_pause -= us_elapsed;
			midi_state.time += us_elapsed;
		} else {
			// Need to process MIDI events.
			int events_left = midi_state.track->num_events - midi_state.event_index;
			if (events_left > 0) {
				midi_event_type* event = &midi_state.track->events[midi_state.event_index++];
				process_midi_event(event);

				// Need to look ahead: must delay processing of the next event, if there is a pause.
				if (events_left > 1) {
					midi_event_type* next_event = &midi_state.track->events[midi_state.event_index];
					if (next_event->delta_time != 0) {
						int us_needed = (us_per_beat / ticks_per_beat) * next_event->delta_time;
						midi_state.time_to_next_pause += us_needed;
						printf("delaying %d ticks = %.3f s\n", next_event->delta_time, (us_needed / 1e6f));
					}
				}
			} else {
				// reached the last event. Fill the remaining samples with silence.
				SDL_memset(stream, 0, frames_needed * 4);
				is_playing = 0;
			}
		}
	}
}

int main(int argc, char** argv) {
	if (argc < 2) {
		printf("Usage: %s [midi-file]\n", argv[0]);
		panic_and_exit();
	}

	file_buffer_type* instruments_file = load_file_into_memory("res1.bin");
	if (!instruments_file) {
		printf("Missing instruments file\n");
		panic_and_exit();
	} else {
		num_instruments = instruments_file->data[0];
		instruments = (instrument_type*) &instruments_file->data[1];
	}

	file_buffer_type* midi_file = load_file_into_memory(argv[1]);
	if (!midi_file) {
		panic_and_exit();
	} else {
		if (!parse_midi((midi_raw_chunk_type*) midi_file->data, &parsed_midi)) {
			panic_and_exit();
		}
//		print_midi_events(&parsed_midi);
	}

	midi_state.track = &parsed_midi.tracks[0];
	us_per_beat = 60000;

	SDL_SetHint("SDL_HINT_WINDOWS_DISABLE_THREAD_NAMING", "1"); // Fix weird gdb debugging crash on Windows.
	if (SDL_Init(SDL_INIT_AUDIO) < 0) {
		sdl_panic("SDL_Init");
	}

	SDL_AudioSpec desired_spec =
			{.freq = 44100, .format = AUDIO_S16, .channels = 2, .samples = 2048, .callback = opl_audio_callback};
//			{.freq = 44100, .format = AUDIO_S16, .channels = 2, .samples = 2048, .callback = sine_audio_callback};

	if (SDL_OpenAudio(&desired_spec, NULL) < 0) {
		sdl_panic("SDL_OpenAudio");
	}
	mix_buffer = malloc(MIX_BUFFER_SIZE);

	opl_reset(desired_spec.freq);

#if 0
	// Sample OPL register values, to test getting any sound at all from the OPL emulator.
	// Source/credit: http://www.shipbrook.net/jeff/sb.html
	// (Need to set up the audio callback differently to get sound out of this, though.)
	opl_write_reg(0x20, 0x01); // Set the modulator's multiple to 1
	opl_write_reg(0x40, 0x10); // Set the modulator's level to about 40 dB
	opl_write_reg(0x60, 0xF0); // Modulator attack: quick; decay; long
	opl_write_reg(0x80, 0x77); // Modulator sustain: medium; release: medium
	opl_write_reg(0x23, 0x01); // Set the carrier's multiple to 1
	opl_write_reg(0x43, 0x00); // Set the carrier to maximum volume (about 47 dB)
	opl_write_reg(0x63, 0xF0); // Carrier attack: quick; decay: long
	opl_write_reg(0x83, 0x77); // Carrier sustain: medium; release: medium
	opl_write_reg(0xA0, 0x98); // Set voice frequency's LSB (it'll be a D#)
	opl_write_reg(0xB0, 0x31); // Turn the voice on; set the octave and freq MSB
#endif

	for (int voice = 0; voice < 9; ++voice) {
		opl_write_instrument(&instruments[0], voice);
	}


	mixing_freq = desired_spec.freq;
	is_playing = 1;
	audio_frames_left = 100000;
	time_division = 1.0 / 44100.0;
	sine_frequency = 2000.0;
	SDL_PauseAudio(0);



	while(is_playing) {
		SDL_Delay(100);
	}


//	midi_iterate_events(&midi_state);

	return 0;
}
